from greyhorse.app.abc.operators import AssignOperator, Operator
from greyhorse.app.abc.providers import FactoryError, FactoryProvider, SharedProvider
from greyhorse.app.boxes import SharedRefBox
from greyhorse.app.entities.controllers import SyncController, operator
from greyhorse.app.entities.services import SyncService, provider
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Result

from .common.functional import FunctionalOperator


class FunctionalOperatorService(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._res = Nothing

    def get_res(self) -> Maybe[FunctionalOperator]:
        return self._res

    def set_res(self, res: Maybe[FunctionalOperator]) -> None:
        self._res = res

    @provider(SharedProvider[FunctionalOperator])
    def create_shared_prov(self) -> SharedProvider[FunctionalOperator] | None:
        if not self._res:
            return None
        return SharedRefBox(self.get_res, lambda v: v)

    @provider(FactoryProvider[FunctionalOperator])
    def create_factory_prov(self) -> Result[FunctionalOperator, FactoryError] | None:
        if not self._res:
            return None
        yield self._res.unwrap()


class FunctionalOperatorCtrl(SyncController):
    def __init__(self, svc: FunctionalOperatorService) -> None:
        super().__init__()
        self._svc = svc

    @operator(FunctionalOperator)
    def create_op(self) -> Operator[FunctionalOperator]:
        return AssignOperator[FunctionalOperator](self._svc.get_res, self._svc.set_res)



from greyhorse.app.schemas.components import ComponentConf, ModuleConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from ..common.functional import FunctionalOperator, FunctionalOpProvider
from ..common.resources import (
    DictCtxProvider,
    DictMutCtxProvider,
    DictResContext,
    MutDictResContext,
)
from ..components.functional import DictOperatorCtrl, DictOperatorService1, DictOperatorService2
from ..components.resources import DictProviderService


def __init__() -> ModuleConf:  # noqa: N807
    return ModuleConf(
        enabled=True,
        operators=[FunctionalOperator],
        components={
            'domain': ComponentConf(
                enabled=True,
                providers=[DictCtxProvider, DictMutCtxProvider],
                services=[SvcConf(type=DictProviderService)],
            ),
            'app': ComponentConf(
                enabled=True,
                providers=[FunctionalOpProvider],
                operators=[DictResContext, MutDictResContext],
                services=[
                    SvcConf(type=DictOperatorService1, resources=[DictResContext]),
                    SvcConf(type=DictOperatorService2, resources=[MutDictResContext]),
                ],
                controllers=[CtrlConf(type=DictOperatorCtrl)],
            ),
        },
    )



import asyncio
import time

import pytest
from greyhorse.app.runtime import Runtime


def sync_fn(runtime: Runtime, counter: int) -> list[int]:
    if counter == 0:
        return [0]

    res = runtime.invoke_sync(async_fn, runtime, counter - 1)
    return [*res, counter]


async def async_fn(runtime: Runtime, counter: int) -> list[int]:
    if counter == 0:
        return [0]

    res = await runtime.invoke_async(sync_fn, runtime, counter - 1)
    return [*res, counter]


def test_runtime_sync() -> None:
    runtime = Runtime()

    asyncio.set_event_loop(runtime.loop)

    counter = 100

    res = runtime.invoke_sync(async_fn, runtime, counter)
    assert list(range(0, counter + 1)) == res

    runtime.start()

    res = runtime.invoke_sync(async_fn, runtime, counter)
    assert list(range(0, counter + 1)) == res

    runtime.stop()


@pytest.mark.asyncio
async def test_runtime_async() -> None:
    runtime = Runtime()

    counter = 100

    res = await runtime.invoke_async(async_fn, runtime, counter)
    assert list(range(0, counter + 1)) == res

    runtime.start()

    res = await runtime.invoke_async(async_fn, runtime, counter)
    assert list(range(0, counter + 1)) == res

    runtime.stop()


if __name__ == '__main__':
    t1 = time.perf_counter()
    test_runtime_sync()
    t2 = time.perf_counter()
    print(f'Sync runtime takes {t2 - t1:.03} secs')

    t1 = time.perf_counter()
    asyncio.run(test_runtime_async())
    t2 = time.perf_counter()
    print(f'Async runtime takes {t2 - t1:.03} secs')



import asyncio
import inspect
import threading
from asyncio import Future as AsyncFuture
from asyncio import iscoroutine, iscoroutinefunction
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from queue import Empty, Queue
from typing import Any


def is_awaitable(f: object) -> bool:
    while isinstance(f, partial):
        f = f.func
    return iscoroutinefunction(f) or inspect.isawaitable(f)


@dataclass(slots=True, frozen=True)
class _Task:
    func: Callable[[...], Any]
    args: tuple
    kwargs: dict[str, Any]
    waiter: AsyncFuture


class Runtime:
    __slots__ = ('_loop', '_thread', '_counter', '_tasks', '_sync_ctx')

    def __init__(self) -> None:
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(name='Greyhorse runtime', target=self._loop.run_forever)
        self._counter = 0
        self._tasks = Queue()
        self._sync_ctx = False

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def active(self) -> bool:
        return self._thread.is_alive()

    def start(self) -> None:
        if self._counter == 0:
            self._thread.start()
        self._counter += 1

    def stop(self) -> None:
        if self._counter == 1:
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join()
            self._loop.close()

        self._counter = max(self._counter - 1, 0)

    def invoke_sync[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if self.active:
                future = asyncio.run_coroutine_threadsafe(func(*args, **kwargs), self._loop)
                self._wait_for_async_future(future)
                return future.result()
            return asyncio.run(func(*args, **kwargs))

        return func(*args, **kwargs)

    async def invoke_async[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if iscoroutine(func):
                return await func  # type: ignore
            return await func(*args, **kwargs)

        if self._sync_ctx:
            future = self._loop.create_future()
            task = _Task(func=func, args=args, kwargs=kwargs, waiter=future)
            self._tasks.put(task)
            return await future

        return await asyncio.to_thread(func, *args, **kwargs)

    def _wait_for_async_future(self, future: AsyncFuture) -> None:
        self._sync_ctx = not future.done()
        while self._sync_ctx:
            try:
                while task := self._tasks.get(timeout=0.01):
                    try:
                        res = task.func(*task.args, **task.kwargs)
                        self._loop.call_soon_threadsafe(task.waiter.set_result, res)

                    except Exception as e:
                        self._loop.call_soon_threadsafe(task.waiter.set_exception, e)

                    self._tasks.task_done()

            except Empty:
                self._sync_ctx = not future.done()


instance = Runtime()



from __future__ import annotations

from greyhorse.app.abc.providers import FactoryProvider, SharedProvider
from greyhorse.app.boxes import PermanentForwardBox
from greyhorse.app.builders.module import ModuleBuilder
from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from .common.functional import FunctionalOperator
from .root import FunctionalOperatorCtrl, FunctionalOperatorService


def __init__() -> ModuleConf:  # noqa: N807
    return ModuleConf(
        enabled=True,
        operators=[FunctionalOperator],
        providers=[FactoryProvider[FunctionalOperator]],
        components={
            'dict': ModuleComponentConf(
                enabled=True,
                path='..module.main',
                providers=[
                    SharedProvider[FunctionalOperator],
                    FactoryProvider[FunctionalOperator],
                ],
                services=[SvcConf(type=FunctionalOperatorService)],
                controllers=[CtrlConf(type=FunctionalOperatorCtrl)],
            )
        },
    )


def test_module() -> None:
    module_conf = __init__()

    module_builder = ModuleBuilder(module_conf, 'tests')
    res = module_builder.create_pass()
    assert res.is_ok()
    module = res.unwrap()

    op_box = PermanentForwardBox[FunctionalOperator]()

    assert module.add_operator(op_box)

    res = module.setup()
    assert res.is_ok()

    prov = module.get_provider(FactoryProvider[FunctionalOperator]).unwrap()

    print('OK')

    op = prov.create().unwrap()

    res = op.add_number(123)
    assert res.is_ok()

    res = op.get_number()
    assert res.is_ok()
    assert res.unwrap() == 123

    res = op.remove_number()
    assert res.is_ok()
    assert res.unwrap() is True

    res = op.get_number()
    assert res.is_err()
    assert res.unwrap_err() == 'Number is not initialized'

    prov.destroy(op)

    res = module.teardown()
    assert res.is_ok()



from collections.abc import Callable, Iterable
from functools import partial
from typing import Any, cast, get_type_hints, override

from pydantic import BaseModel

from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Ok, Result

from ..abc.collectors import MutNamedCollector, NamedCollector
from ..abc.controllers import Controller, ControllerError, OperatorMember
from ..abc.operators import AssignOperator


class ResConf(BaseModel, frozen=True):
    type: type
    name: str | None = None
    required: bool = True


def operator(resource_type: type) -> Callable[[classmethod], classmethod]:
    def decorator(func: classmethod) -> classmethod:
        hints = get_type_hints(func, include_extras=True)
        ret_type = hints.pop('return', None)
        func.__operator__ = OperatorMember(
            resource_type, method=func, params=hints, ret_type=ret_type
        )
        return func

    return decorator


class SyncController(Controller):
    @override
    def setup(self, collector: NamedCollector[type, Any]) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    def teardown(
        self, collector: MutNamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        return Ok(True)


class AsyncController(Controller):
    @override
    async def setup(
        self, collector: NamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    async def teardown(
        self, collector: MutNamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        return Ok(True)


class ResourceController(SyncController):
    def __init__(self, resources: Iterable[ResConf]) -> None:
        super().__init__()
        self._res_types = resources
        self._values: list[Maybe[Any]] = [Nothing for _ in self._res_types]
        self._compile_operator_methods()

    def _getter(self, idx: int) -> Maybe[Any]:
        if 0 <= idx < len(self._values):
            return self._values[idx]
        return Nothing

    def _setter(self, idx: int, value: Maybe[Any]) -> None:
        if 0 <= idx < len(self._values):
            self._values[idx] = value

    @override
    def setup(self, collector: NamedCollector[type, Any]) -> Result[bool, ControllerError]:
        for idx, res_conf in enumerate(self._res_types):
            if (
                not self._values[idx]
                .map(
                    lambda value, res_conf=res_conf: collector.add(
                        res_conf.type, value, name=res_conf.name
                    )
                )
                .unwrap_or(False)
                and res_conf.required
            ):
                return ControllerError.NoSuchResource(
                    name=f'{res_conf.type.__name__}'
                ).to_result()

        return super().setup(collector)

    @override
    def teardown(
        self, collector: MutNamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        for idx, res_conf in enumerate(self._res_types):
            if (
                not self._values[idx]
                .map(
                    lambda value, res_conf=res_conf: collector.remove(
                        res_conf.type, value, name=res_conf.name
                    )
                )
                .unwrap_or(False)
                and res_conf.required
            ):
                return ControllerError.NoSuchResource(
                    name=f'{res_conf.type.__name__}'
                ).to_result()

        return super().teardown(collector)

    def _compile_operator_methods(self) -> None:
        for idx, res_conf in enumerate(self._res_types):
            method_name = f'_create_op_{res_conf.type.__name__}'

            def func(self, idx=idx, res_conf=res_conf):  # noqa
                return AssignOperator[res_conf.type](
                    partial(self._getter, idx), partial(self._setter, idx)
                )

            func.__qualname__ = f'{self.__class__.__name__}.{method_name}'
            func = cast(classmethod, partial(func, self=self))

            member = OperatorMember(res_conf.type, method=func, params={})
            self._operator_members[res_conf.type] = member



from typing import Any, override

from greyhorse.app.abc.collectors import MutNamedCollector, NamedCollector
from greyhorse.app.abc.controllers import ControllerError
from greyhorse.app.abc.operators import AssignOperator, Operator
from greyhorse.app.abc.providers import FactoryError, ForwardProvider
from greyhorse.app.abc.selectors import NamedListSelector, NamedSelector
from greyhorse.app.abc.services import ServiceError, ServiceState
from greyhorse.app.boxes import ForwardBox
from greyhorse.app.entities.controllers import SyncController, operator
from greyhorse.app.entities.services import SyncService, provider
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Ok, Result

from ..common.functional import FunctionalOperator, FunctionalOpProvider
from ..common.resources import DictResContext, MutDictResContext


class FunctionalOperatorImpl(FunctionalOperator):
    def __init__(self, ctx: DictResContext, mut_ctx: MutDictResContext) -> None:
        self.ctx = ctx
        self.mut_ctx = mut_ctx

    @override
    def add_number(self, value: int) -> Result[None, str]:
        with self.mut_ctx as ctx:
            ctx['number'] = value
            self.mut_ctx.apply()

        return Ok()

    @override
    def get_number(self) -> Result[int, str]:
        with self.ctx as ctx:
            value = Maybe(ctx.get('number'))

        return value.ok_or('Number is not initialized')

    @override
    def remove_number(self) -> Result[bool, str]:
        with self.mut_ctx as ctx:
            value = Maybe(ctx.pop('number', None))
            self.mut_ctx.apply()

        return value.ok_or('Number is not initialized').map(lambda _: True)


class DictOperatorService1(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._res1 = ForwardBox[DictResContext]()

    @override
    def setup(
        self, res: Maybe[DictResContext], selector: NamedSelector[type, Any]
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        self._res1.accept(res.unwrap())
        return super().setup()

    @override
    def teardown(
        self, res: Maybe[DictResContext], selector: NamedSelector[type, Any]
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        if not selector.has(DictResContext):
            return ServiceError.NoSuchResource(name='DictResContext').to_result()
        self._res1.revoke()
        return super().teardown()

    def start(self) -> None:
        self._switch_to_active(True)

    def stop(self) -> None:
        self._switch_to_active(False)

    @provider(ForwardProvider[DictResContext])
    def create_prov(self) -> ForwardProvider[DictResContext]:
        return self._res1


class DictOperatorService2(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._res2: MutDictResContext | None = None

    @override
    def setup(
        self, res: Maybe[MutDictResContext], list_selector: NamedListSelector[type, Any]
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='MutDictResContext').to_result()
        if not list_selector.has(MutDictResContext):
            return ServiceError.NoSuchResource(name='MutDictResContext').to_result()
        self._res2 = res.unwrap()
        return super().setup()

    @override
    def teardown(
        self, res: Maybe[MutDictResContext], list_selector: NamedListSelector[type, Any]
    ) -> Result[ServiceState, ServiceError]:
        if not res:
            return ServiceError.NoSuchResource(name='MutDictResContext').to_result()
        if not list_selector.has(MutDictResContext):
            return ServiceError.NoSuchResource(name='MutDictResContext').to_result()
        self._res2 = None
        return super().teardown()

    def start(self) -> None:
        self._switch_to_active(True)

    def stop(self) -> None:
        self._switch_to_active(False)

    @provider(FunctionalOpProvider)
    def create_prov(
        self, dependency: ForwardProvider[DictResContext]
    ) -> Result[FunctionalOpProvider, FactoryError] | None:
        if not dependency:
            return FactoryError.InsufficientDeps(name='DictResContext').to_result()
        if not self._res2:
            return None
        res1 = dependency.take().unwrap()
        yield FunctionalOperatorImpl(res1, self._res2)
        dependency.drop(res1)


class DictOperatorCtrl(SyncController):
    def __init__(self) -> None:
        super().__init__()
        self._a: Maybe[DictResContext] = Nothing
        self._b: Maybe[MutDictResContext] = Nothing

    def _setter1(self, value: Maybe[DictResContext]) -> None:
        self._a = value

    def _setter2(self, value: Maybe[MutDictResContext]) -> None:
        self._b = value

    @override
    def setup(self, collector: NamedCollector[type, Any]) -> Result[bool, ControllerError]:
        if not self._a:
            return ControllerError.NoSuchResource(name='DictResContext').to_result()
        if not self._b:
            return ControllerError.NoSuchResource(name='MutDictResContext').to_result()
        collector.add(DictResContext, self._a.unwrap())
        collector.add(MutDictResContext, self._b.unwrap())
        return super().setup(collector)

    @override
    def teardown(
        self, collector: MutNamedCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        if not self._a:
            return ControllerError.NoSuchResource(name='DictResContext').to_result()
        if not self._b:
            return ControllerError.NoSuchResource(name='MutDictResContext').to_result()
        collector.remove(DictResContext, self._a.unwrap())
        collector.remove(MutDictResContext, self._b.unwrap())
        return super().teardown(collector)

    @operator(DictResContext)
    def create_op1(self) -> Operator[DictResContext]:
        return AssignOperator[DictResContext](lambda: self._a, self._setter1)

    @operator(MutDictResContext)
    def create_op2(self) -> Operator[MutDictResContext]:
        return AssignOperator[MutDictResContext](lambda: self._b, self._setter2)



from typing import Any

from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.visitor import Visitor
from greyhorse.app.entities.components import Component
from greyhorse.app.private.res_manager import ResourceManager
from greyhorse.app.registries import MutDictRegistry, MutNamedDictRegistry
from greyhorse.app.schemas.components import ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.result import Ok, Result


class ModuleError(Error):
    namespace = 'greyhorse.app.module'

    Component = ErrorCase(
        msg='{path}: Component error in module, details: "{details}"', path=str, details=str
    )

    Resource = ErrorCase(
        msg='{path}: Resource error in module: "{details}"', path=str, details=str
    )


class Module:
    def __init__(self, path: str, conf: ModuleConf, components: list[Component]) -> None:
        self._conf = conf
        self._path = path
        self._rm = ResourceManager()

        self._operators: list[Operator] = []

        self._resources = MutNamedDictRegistry[type, Any]()
        self._providers = MutDictRegistry[type[Provider], Provider]()
        self._components: dict[str, Component] = {c.name: c for c in components}

    @property
    def path(self) -> str:
        return self._path

    @property
    def conf(self) -> ModuleConf:
        return self._conf

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_module(self)
        for comp in self._components.values():
            comp.accept_visitor(visitor)
        visitor.finish_module(self)

    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]:
        if prov_type in self._conf.providers:
            return (
                self._rm.find_provider(prov_type, self._providers).map(Just).unwrap_or(Nothing)
            )
        return Nothing

    def add_provider[T](self, prov_type: type[Provider[T]], provider: Provider[T]) -> bool:
        if prov_type in self._conf.provider_claims:
            return self._providers.add(prov_type, provider)
        return False

    def remove_provider[T](self, prov_type: type[Provider[T]]) -> bool:
        if prov_type in self._conf.provider_claims:
            return self._providers.remove(prov_type)
        return False

    def add_resource(self, res_type: type, resource: Any, name: str | None = None) -> bool:  # noqa: ANN401
        if res_type in self._conf.resource_claims:
            return self._resources.add(res_type, resource, name=name)
        return False

    def remove_resource(self, res_type: type, name: str | None = None) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.remove(res_type, name=name)
        return False

    def add_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.operators:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.append(operator)
                return True
        return False

    def remove_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.operators:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.remove(operator)
                return True
        return False

    def setup(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module setup'.format(path=self._path))

        for component in self._components.values():
            comp_conf = self._conf.components[component.name]

            for res_type in comp_conf.resource_claims:
                if res := self._resources.get(res_type).unwrap_or_none():
                    component.add_resource(res_type, res)

            # XXX: component providers
            # for prov_type in prov_conf.providers:
            #     for _, prov in self._providers.items(
            #         lambda t, pt=prov_type: issubclass(t, pt)
            #     ):
            #         component.add_provider(prov_type, prov)

            if not (
                res := component.create().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for res_type in comp_conf.operators:
                for op in component.get_operators(res_type):  # type: ignore
                    if not (
                        res := self._rm.setup_resource(op, self._providers).map_err(
                            lambda e: ModuleError.Resource(path=self._path, details=e.message)
                        )
                    ):
                        return res  # type: ignore

            if not (
                res := component.setup().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for prov_type in comp_conf.providers:
                if prov := component.get_provider(prov_type).unwrap_or_none():
                    self._providers.add(prov_type, prov)

        for op in self._operators:
            if not (
                res := self._rm.setup_resource(op, self._providers).map_err(
                    lambda e: ModuleError.Resource(path=self._path, details=e.message)
                )
            ):
                return res  # type: ignore

        logger.info('{path}: Module setup successful'.format(path=self._path))
        return Ok()

    def teardown(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module teardown'.format(path=self._path))

        for op in reversed(self._operators):
            if not (
                res := self._rm.teardown_resource(op).map_err(
                    lambda e: ModuleError.Resource(path=self._path, details=e.message)
                )
            ):
                return res  # type: ignore

        for component in reversed(self._components.values()):
            comp_conf = self._conf.components[component.name]

            for prov_type in reversed(comp_conf.providers):
                self._providers.remove(prov_type)

            if not (
                res := component.teardown().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for res_type in reversed(comp_conf.operators):
                for op in component.get_operators(res_type):  # type: ignore
                    if not (
                        res := self._rm.teardown_resource(op).map_err(
                            lambda e: ModuleError.Resource(path=self._path, details=e.message)
                        )
                    ):
                        return res  # type: ignore

            if not (
                res := component.destroy().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            # XXX: component providers
            # for prov_type in reversed(comp_conf.provider_grants):
            #     component.remove_provider(prov_type)

            for res_type in reversed(comp_conf.resource_claims):
                component.remove_resource(res_type)

        if not (
            res := self._rm.teardown().map_err(
                lambda e: ModuleError.Resource(path=self._path, details=e.message)
            )
        ):
            return res

        logger.info('{path}: Module teardown successful'.format(path=self._path))
        return Ok()



# mypy: warn_no_return=false,disable_error_code="arg-type,valid-type"
from __future__ import annotations

from dataclasses import fields as dataclass_fields
from typing import TYPE_CHECKING, Any, ClassVar, Final, Self

from .enum import Enum, Struct
from .i18n import StaticTranslator

if TYPE_CHECKING:
    from .result import Result


class Error(Enum):
    _tr: ClassVar[StaticTranslator] | None = None
    namespace: ClassVar[str] = ''
    code: Final[str] = ''
    msg: Final[str] = ''

    @property
    def message(self) -> str:
        values = {}

        # noinspection PyDataclass
        for field in dataclass_fields(self):
            if not field.repr or not field.init:
                continue

            values[field.name] = getattr(self, field.name, '')

        if self.code and self._tr is not None:
            msg = self._tr('.'.join([self.namespace, self.code]), default=self.msg)
        else:
            msg = self.msg

        return msg.format(**values)

    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        if 'tr' in kwargs:
            assert isinstance(kwargs['tr'], StaticTranslator)
            cls._tr = kwargs.pop('tr')
        super().__init_subclass__(**kwargs)

    def to_result(self) -> Result[Any, Self]:
        from .result import Err

        return Err(self)


class ErrorCase(Struct):
    __slots__ = ('msg', 'code')

    def _repr_fn(self, instance: Error) -> str:
        return f'{self._base.__name__}:{self._name}("{instance.message}")'



import asyncio
from collections.abc import Callable
from functools import partial

from greyhorse.app.runtime import instance, is_awaitable


def run[T, **P](func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> T:
    asyncio.set_event_loop(instance.loop)

    try:
        instance.start()

        if is_awaitable(func):
            return asyncio.run(func(*args, **kwargs))
        return func(*args, **kwargs)

    finally:
        instance.stop()


def main[T, **P](func: Callable[P, T] | None = None) -> T:
    def decorator[T, **P](func: Callable[P, T]) -> T:
        return partial(run, func)

    if func is None:
        return decorator

    return decorator(func)



import sys
from collections.abc import Callable
from typing import cast

import pydantic

from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Ok, Result
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync


class ModuleLoadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyLoaded = ErrorCase(msg='Module already loaded: "{path}"', path=str)
    WrongImport = ErrorCase(
        msg='Module import error: "{path}", details: "{details}"', path=str, details=str
    )
    Validation = ErrorCase(
        msg='Module validation error: "{path}", details: "{details}"', path=str, details=str
    )
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )
    InvalidConf = ErrorCase(msg='Module conf is invalid: "{path}"', path=str)


class ModuleUnloadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyUnloaded = ErrorCase(msg='Module already unloaded: "{path}"', path=str)
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )


class ModuleLoader:
    def __init__(self) -> None:
        self._injector = ParamsInjector()

    # noinspection PyProtectedMember
    def load_pass(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        if conf._conf is not None:
            return ModuleLoadError.AlreadyLoaded(path=conf.path).to_result()

        if not (res := self._load_module(conf)):
            return res

        conf._conf = res.unwrap()

        if not conf._conf.enabled:
            return ModuleLoadError.Disabled(path=conf.path).to_result()

        return Ok(conf._conf)

    # noinspection PyProtectedMember
    def unload_pass(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        if conf._conf is None:
            return ModuleUnloadError.AlreadyUnloaded(path=conf.path).to_result()

        if not conf._conf.enabled:
            return ModuleUnloadError.Disabled(path=conf.path).to_result()

        return self._unload_module(conf)

    # noinspection PyProtectedMember
    @staticmethod
    def _get_module_package(conf: ModuleComponentConf) -> str:
        if not conf.path.startswith('.'):
            return conf.path

        path = conf.path
        dots_count = 0

        for c in path[1:]:
            if c == '.':
                dots_count += 1
            else:
                path = path[dots_count + 1 :]
                break

        init_path = conf._init_path
        dots_count = min(dots_count, len(init_path))
        init_path = init_path[0 : len(init_path) - dots_count]
        return '.'.join([*init_path, path])

    def _load_module(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to load module "{path}"'.format(path=module_path))

        try:
            func = cast(Callable[[...], ModuleConf], import_path(f'{module_path}:__init__'))

        except (ImportError, AttributeError) as e:
            error = ModuleLoadError.WrongImport(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        injected_args = self._injector(func, values=conf.args)

        try:
            res = invoke_sync(func, *injected_args.args, **injected_args.kwargs)

        except pydantic.ValidationError as e:
            error = ModuleLoadError.Validation(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()
        except Exception as e:
            error = ModuleLoadError.Unexpected(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        if not isinstance(res, ModuleConf):
            error = ModuleLoadError.InvalidConf(path=module_path)
            logger.error(error.message)
            return error.to_result()

        logger.info(
            'ModuleLoader: Module "{path}" loaded successfully'.format(path=module_path)
        )
        return Ok(res)

    # noinspection PyProtectedMember
    def _unload_module(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to unload module "{path}"'.format(path=module_path))

        try:
            func = cast(Callable[[...], None], import_path(f'{module_path}:__fini__'))

        except (ImportError, AttributeError):
            pass
        else:
            injected_args = self._injector(func, types={ModuleConf: conf._conf})

            try:
                invoke_sync(func, *injected_args.args, **injected_args.kwargs)

            except Exception as e:
                error = ModuleUnloadError.Unexpected(path=module_path, details=str(e))
                logger.error(error.message)
                return error.to_result()

        conf, conf._conf = conf._conf, None
        del conf
        del sys.modules[module_path]

        logger.info(
            'ModuleLoader: Module "{path}" unloaded successfully'.format(path=module_path)
        )
        return Ok()



import inspect
from collections.abc import Callable, Mapping
from typing import Any

from greyhorse.maybe import Maybe

TypeProviderFactory = Callable[[str, ...], Any]


class ParamsInjector:
    def __init__(self, type_providers: TypeProviderFactory | None = None) -> None:
        self._types_providers: dict[type, TypeProviderFactory | Any] = type_providers or {}

    def get_type_provider(self, t: type) -> Maybe[TypeProviderFactory | object]:
        return Maybe(self._types_providers.get(t))

    def add_type_provider(self, t: type, p: TypeProviderFactory | object) -> None:
        self._types_providers[t] = p

    def remove_type_provider(self, t: type) -> None:
        self._types_providers.pop(t, None)

    def __call__(
        self,
        func: Callable[[...], ...],
        values: Mapping[str, Any] | None = None,
        types: Mapping[type, Any] | None = None,
    ) -> inspect.BoundArguments:
        values = values or {}
        types = types or {}

        sig = inspect.signature(func, eval_str=True)
        args = {}

        for k, v in sig.parameters.items():
            if k == 'values':
                args[k] = values
            elif v.annotation in types:
                args[k] = types[v.annotation]
            elif k in values:
                args[k] = values[k]
            elif v.annotation in self._types_providers:
                args[k] = self._types_providers[v.annotation]

        args = sig.bind_partial(**args)
        args.apply_defaults()
        return args
